/* BBVHM - maximize a window vertically or horizontally (can be undone)

This program can be redistributed and/or modified freely.

Copyright (C) 2004 kana <nicht AT s8 DOT xrea DOT com>
*/

#include <stdlib.h>

#include <BBApi-C.h>

#define NUMBER_OF(a) (sizeof(a) / sizeof((a)[0]))




static const char BBVHM_NAME[] = "BBVHM";
static const char BBVHM_VERSION[] = "0.0.0";
static const char BBVHM_AUTHOR[] = "kana";
static const char BBVHM_RELEASE[] = "2004-11-01";
static const char BBVHM_LINK[] = "http://nicht.s8.xrea.com/";
static const char BBVHM_EMAIL[] = "nicht AT s8 DOT xrea DOT con";

static UINT BBVHM_MESSAGES[] = {BB_BROADCAST, 0};

enum {
	VERTICAL,
	HORIZONTAL
};




typedef struct {
	HWND hwnd;
	RECT rect;
} InfoEntry;

typedef struct {
	int size;
	InfoEntry* entry;
} Info;


struct {
	HWND plugin_window;
	Info info;
} G;








static void
error(const char* format, ...)
{
	char buf[80*25];
	va_list va;

	va_start(va, format);
	vsnprintf(buf, NUMBER_OF(buf), format, va);
	va_end(va);

	MessageBox( G.plugin_window, buf, BBVHM_NAME,
	            MB_OK | MB_ICONERROR
	              | MB_SETFOREGROUND | MB_TASKMODAL | MB_TOPMOST );
}




static void
Info_Initialize(Info* self)
{
	self->size = 0;
	self->entry = NULL;
}

static void
Info_Finalize(Info* self)
{
	free(self->entry);
}


static int
Info_Find(Info* self, HWND hwnd)
{
	int i;

	for (i = 0; i < self->size; i++) {
		if (hwnd == self->entry[i].hwnd)
			return i;
	}

	return -1;
}

static const RECT*
Info_Get(Info* self, HWND hwnd)
{
	int index;

	index = Info_Find(self, hwnd);
	if (index == -1)
		return NULL;

	return &(self->entry[index].rect);
}

static void
Info_Add(Info* self, HWND hwnd)
{
	InfoEntry* t;

	t = realloc(self->entry, sizeof(t[0]) * (self->size + 1));
	if (t == NULL) {
		error("Out of memory");
		return;
	}

	self->entry = t;
	self->entry[self->size].hwnd = hwnd;
	GetWindowRect(hwnd, &(self->entry[self->size].rect));
	self->size++;
}

static void
Info_Remove(Info* self, HWND hwnd)
{
	int i;
	int j;

	j = Info_Find(self, hwnd);
	if (j == -1)
		return;

	for (i = j + 1; i < self->size; i++)
		self->entry[i - 1] = self->entry[i];

	self->size--;

	/* Should I call realloc() here? */
}








static void
maximize(int direction)
{
	HWND hwnd;
	const RECT* rect;

	hwnd = GetForegroundWindow();
	hwnd = GetAncestor(hwnd, GA_ROOTOWNER);
	hwnd = GetLastActivePopup(hwnd);

	if (hwnd == NULL)
		return;

	rect = Info_Get(&(G.info), hwnd);

	if (rect == NULL) {  /* the window is not maximized */
		Info_Add(&(G.info), hwnd);

		SendMessage( GetBBWnd(),
		            ( direction == VERTICAL ? BB_WINDOWGROWHEIGHT
		                                    : BB_WINDOWGROWWIDTH ),
		            (WPARAM)hwnd,
		            0 );
	} else {  /* the window is already maximized */
		MoveWindow( hwnd,
		            rect->left,
		            rect->top,
		            rect->right - rect->left,
		            rect->bottom - rect->top,
		            TRUE );
		Info_Remove(&(G.info), hwnd);
	}
}




static void
interpret_broam(const char* broam)
{
	if (!strcmp(broam, "@BBVHM.maximize_vertical"))
		maximize(VERTICAL);
	else if (!strcmp(broam, "@BBVHM.maximize_horizontal"))
		maximize(HORIZONTAL);
	else
		;  /* nop */
}




static LRESULT CALLBACK
window_procedure(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg) {
	default:
		break;


	case BB_BROADCAST:
		interpret_broam((const char*)lp);
		break;


	case WM_CREATE:
		SendMessage( GetBBWnd(), BB_REGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)BBVHM_MESSAGES );
		break;

	case WM_DESTROY:
		SendMessage( GetBBWnd(), BB_UNREGISTERMESSAGE,
		             (WPARAM)hwnd, (LPARAM)BBVHM_MESSAGES );
		break;
	}

	return DefWindowProc(hwnd, msg, wp, lp);
}








DLL_EXPORT int
beginPlugin(HINSTANCE hplugininstance)
{
	WNDCLASS wc;

	Info_Initialize(&(G.info));

	ZeroMemory(&wc, sizeof(wc));
	wc.lpszClassName = BBVHM_NAME;
	wc.lpfnWndProc = window_procedure;
	wc.hInstance = hplugininstance;
	if (!RegisterClass(&wc)) {
		error("RegisterClass failed");
		goto E_REGISTER_CLASS;
	}

	G.plugin_window = CreateWindowEx( WS_EX_TOOLWINDOW,
	                           BBVHM_NAME,
	                           NULL,
	                           WS_DISABLED,
	                           0, 0, 0, 0,
	                           NULL,
	                           NULL,
	                           hplugininstance,
	                           NULL );
	if (G.plugin_window == NULL) {
		error("CreateWindowEx failed");
		goto E_CREATE_WINDOW_EX;
	}

	return 0;


E_CREATE_WINDOW_EX:
	UnregisterClass(BBVHM_NAME, hplugininstance);
E_REGISTER_CLASS:
	return !0;
}




DLL_EXPORT void
endPlugin(HINSTANCE hplugininstance)
{
	DestroyWindow(G.plugin_window);
	UnregisterClass(BBVHM_NAME, hplugininstance);

	Info_Finalize(&(G.info));
}




DLL_EXPORT LPCSTR
pluginInfo(int field)
{
	switch (field) {
	case PLUGIN_NAME:    return BBVHM_NAME;
	case PLUGIN_VERSION: return BBVHM_VERSION;
	case PLUGIN_AUTHOR:  return BBVHM_AUTHOR;
	case PLUGIN_RELEASE: return BBVHM_RELEASE;
	case PLUGIN_LINK:    return BBVHM_LINK;
	case PLUGIN_EMAIL:   return BBVHM_EMAIL;
	default:             return BBVHM_NAME;
	}
}




/* __END__ */
